from ontolutils import namespaces, urirefs
from .variable import Variable


@namespaces(pims="http://www.molmod.info/semantics/pims-ii.ttl#")
@urirefs(Property='pims:Property')
class Property(Variable):
    """Pydantic implementation of pims:Property"""



import logging
import unittest

import rdflib

from ontolutils.namespacelib import QUDT_UNIT

LOG_LEVEL = logging.DEBUG


class TestDoc(unittest.TestCase):

    def test_doc1(self):
        self.assertEqual(
            rdflib.URIRef('http://qudt.org/vocab/unit/M-PER-SEC'),
            QUDT_UNIT.M_PER_SEC)



import logging
import unittest

from ontolutils import Thing, namespaces, urirefs
from ontolutils.classes import decorator

LOG_LEVEL = logging.DEBUG


class TestDecorator(unittest.TestCase):

    def test__is_http_url(self):
        self.assertFalse(decorator._is_http_url(123))
        self.assertFalse(decorator._is_http_url('123'))
        self.assertTrue(decorator._is_http_url('http://example.com'))

    def test_namespaces(self):
        with self.assertRaises(ValueError):
            @namespaces(invalid='Not_a_URI')
            class MyThing(Thing):
                """MyTHing"""

    def test_urirefs(self):
        @urirefs(name='http://example.com/name')
        class MyThing(Thing):
            """MyTHing"""
            name: str

        @namespaces(ex='http://example.com/')
        @urirefs(name='ex:name')
        class MyThing(Thing):
            """MyTHing"""
            name: str

        with self.assertRaises(TypeError):
            @urirefs(name=1.4)
            class MyThing(Thing):
                """MyTHing"""
                name: str



from typing import Union, Optional, List

from pydantic import HttpUrl, field_validator, Field
from pydantic import model_validator

from ontolutils import Thing, namespaces, urirefs, as_id, LangString
from ontolutils.ex.prov import Organization, Person
from ontolutils.typing import ResourceType


@namespaces(schema="https://schema.org/")
@urirefs(
    Project='schema:Project',
    name='schema:name',
    identifier='schema:identifier',
    funder='schema:funder'
)
class Project(Thing):
    """Implementation of schema:Project"""
    name: Optional[Union[LangString, List[LangString]]] = Field(default=None)
    identifier: Optional[ResourceType] = Field(default=None)
    funder: Optional[Union[Person, Organization]] = Field(default=None)

    @model_validator(mode="before")
    def _change_id(self):
        return as_id(self, "identifier")

    @field_validator('identifier', mode='before')
    @classmethod
    def _identifier(cls, identifier):
        HttpUrl(identifier)
        return str(identifier)


@urirefs(ResearchProject='schema:ResearchProject')
class ResearchProject(Project):
    """Pydantic Model for schema:ResearchProject

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    tbd
    """

    def _repr_html_(self) -> str:
        """Returns the HTML representation of the class"""
        return f"{self.__class__.__name__}({self.identifier})"



import abc

from pydantic import BaseModel, ConfigDict

from ..config import PYDANTIC_EXTRA


class ThingModel(BaseModel, abc.ABC, validate_assignment=True):
    """Abstract base model class to be used by model classes used within ontolutils"""

    model_config = ConfigDict(extra=PYDANTIC_EXTRA, populate_by_name=True)

    def __getattr__(self, item):
        for field, meta in self.__class__.model_fields.items():
            if meta.alias == item:
                return getattr(self, field)
        return super().__getattr__(item)

    @abc.abstractmethod
    def _repr_html_(self) -> str:
        """Returns the HTML representation of the class"""



import unittest

from ontolutils.ex.qudt.utils import iri2str
from ontolutils.namespacelib import QUDT_UNIT


class TestQudt(unittest.TestCase):

    def test_iri2str(self):
        str1 = iri2str[str(QUDT_UNIT.M)]
        self.assertEqual(str1, "m")



import logging
import pathlib
import unittest
from typing import List, Union

from pydantic import Field, HttpUrl

from ontolutils import Property
from ontolutils import Thing, urirefs, namespaces, build
from ontolutils import cache
from ontolutils.ex import dcat, prov
from ontolutils.ex.schema import Project, ResearchProject

__this_dir__ = pathlib.Path(__file__).parent

CACHE_DIR = cache.get_cache_dir()

LOG_LEVEL = logging.DEBUG


class TestSchema(unittest.TestCase):

    def testSchemaHTTP(self):
        @namespaces(schema="https://schema.org/")
        @urirefs(SoftwareSourceCode='schema:SoftwareSourceCode',
                 code_repository='schema:codeRepository',
                 application_category='schema:applicationCategory')
        class SoftwareSourceCode(Thing):
            """Pydantic Model for https://schema.org/SoftwareSourceCode"""
            code_repository: Union[HttpUrl, str] = Field(default=None, alias="codeRepository")
            application_category: Union[str, HttpUrl] = Field(default=None, alias="applicationCategory")

        thing = SoftwareSourceCode.from_jsonld(data={
            "@id": "_:N123",
            "@type": "http://schema.org/SoftwareSourceCode",  # note, it is http instead of https!
            "codeRepository": "https://example.com/code",
            "applicationCategory": "https://example.com/category"
        })
        print(thing[0].model_dump_jsonld(indent=2))

    def test_event(self):
        # Event is not defined in the package
        event1 = build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="Event",
            properties=[Property(
                name="about",
                default=None,
                property_type=Union[Thing, List[Thing]]
            )]
        )
        self.assertIsInstance(
            event1(), Thing
        )
        conference1 = event1(
            about=[Thing(label="my conference")]
        )
        serialized1 = conference1.serialize("ttl")

        event2 = build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="Event",
            properties=[dict(
                name="about",
                default=None,
                property_type=Union[Thing, List[Thing]]
            )]
        )
        self.assertIsInstance(
            event2(), Thing
        )
        conference2 = event2(
            about=[Thing(label="my conference")]
        )
        serialized2 = conference2.serialize("ttl")

        expected_serialization = """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <https://schema.org/> .

[] a schema:Event ;
    schema:about [ a owl:Thing ;
            rdfs:label "my conference" ] .

"""
        self.assertEqual(serialized1, expected_serialization)
        self.assertEqual(serialized2, expected_serialization)

    def test_project(self):
        snt_dist = dcat.Distribution(
            downloadURL="https://sandbox.zenodo.org/records/123202/files/Standard_Name_Table_for_the_Property_Descriptions_of_Centrifugal_Fans.jsonld",
            media_type="application/json+ld"
        )

        dataset = dcat.Dataset(
            identifier="https://sandbox.zenodo.org/uploads/123202",
            distribution=snt_dist
        )

        proj = Project(
            identifier='https://example.com/project',
            funder=prov.Organization(name='Funder'),
            usesStandardnameTable=dataset
        )

        self.assertEqual(str(proj.identifier), 'https://example.com/project')
        self.assertIsInstance(proj.funder, prov.Organization)
        self.assertEqual(proj.funder.name, 'Funder')
        self.assertEqual(proj.id, 'https://example.com/project')

    def test_research_project(self):
        proj = ResearchProject(
            identifier='https://example.com/research_project',
            funder=prov.Organization(name='Funder')
        )
        self.assertEqual(proj.id, 'https://example.com/research_project')



import abc
import warnings
from datetime import datetime
from typing import Any
from typing import List, Union
from typing import Optional

from pydantic import HttpUrl, field_validator, Field

from ontolutils import Thing, namespaces, urirefs
from ontolutils import parse_unit, LangString
from ontolutils.ex.pimsii import Variable
from ..prov import Organization
from ..schema import ResearchProject


@namespaces(m4i="http://w3id.org/nfdi4ing/metadata4ing#")
@urirefs(TextVariable='m4i:TextVariable',
         hasStringValue='m4i:hasStringValue')
class TextVariable(Variable):
    """Pydantic Model for http://www.w3.org/ns/prov#Agent

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    hasStringValue: str
        String value
    """
    hasStringValue: Optional[LangString] = Field(alias="has_string_value", default=None)


@namespaces(m4i="http://w3id.org/nfdi4ing/metadata4ing#")
@urirefs(NumericalVariable='m4i:NumericalVariable',
         hasUnit='m4i:hasUnit',
         hasNumericalValue='m4i:hasNumericalValue',
         hasMaximumValue='m4i:hasMaximumValue')
class NumericalVariable(Variable):
    hasUnit: Optional[str] = Field(alias="has_unit", default=None)
    hasNumericalValue: Optional[Union[float, List[float]]] = Field(alias="has_numerical_value", default=None)
    hasMaximumValue: Optional[float] = Field(alias="has_maximum_value", default=None)

    @field_validator("hasUnit", mode='before')
    @classmethod
    def _parse_unit(cls, unit):
        if unit.startswith("http"):
            return str(unit)
        try:
            return parse_unit(unit)
        except KeyError as e:
            warnings.warn(f"Unit '{unit}' could not be parsed to QUDT IRI. This is a process based on a dictionary "
                          f"lookup. Either the unit is wrong or it is not yet included in the dictionary. ")
        return str(unit)


@namespaces(m4i="http://w3id.org/nfdi4ing/metadata4ing#",
            schema="https://schema.org/")
@urirefs(Method='m4i:Method',
         description='schema:description',
         parameter='m4i:hasParameter')
class Method(Thing):
    """Pydantic Model for m4i:M4IProcessingStep

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    description: str
        Description
    parameter: Variable
        Variable(s) used in the method
    """
    description: str = None
    parameter: Union[Variable, List[Variable], NumericalVariable, List[NumericalVariable]] = None

    def add_numerical_variable(self, numerical_variable: Union[dict, "NumericalVariable"]):
        """add numerical variable to tool"""
        if isinstance(numerical_variable, dict):
            # lokaler Import vermeidet zirkulären Import beim Modul-Import
            from ontolutils.ex.m4i import NumericalVariable
            numerical_variable = NumericalVariable(**numerical_variable)
        if self.parameter is None:
            self.parameter = [numerical_variable]
        elif isinstance(self.parameter, list):
            self.parameter.append(numerical_variable)
        else:
            self.parameter = [self.parameter, numerical_variable]


@namespaces(m4i="http://w3id.org/nfdi4ing/metadata4ing#",
            pivmeta="https://matthiasprobst.github.io/pivmeta#",
            obo="http://purl.obolibrary.org/obo/")
@urirefs(Tool='m4i:Tool',
         manufacturer='pivmeta:manufacturer',
         hasParameter='m4i:hasParameter',
         BFO_0000051='obo:BFO_0000051')
class Tool(Thing):
    """Pydantic Model for m4i:ProcessingStep

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    hasParameter: TextVariable or NumericalVariable or list of them
        Text or numerical variable
    """
    hasParameter: Union["TextVariable", "NumericalVariable",
    List[Union["TextVariable", "NumericalVariable"]]] = Field(default=None, alias="parameter")
    manufacturer: Organization = Field(default=None)
    BFO_0000051: Optional[Union[Thing, List[Thing]]] = Field(alias="has_part", default=None)

    @property
    def hasPart(self):
        return self.BFO_0000051

    @hasPart.setter
    def hasPart(self, value):
        self.BFO_0000051 = value

    def add_numerical_variable(self, numerical_variable: Union[dict, NumericalVariable]):
        """add numerical variable to tool"""
        if isinstance(numerical_variable, dict):
            numerical_variable = NumericalVariable(**numerical_variable)
        if self.parameter is None:
            self.hasParameter = [numerical_variable, ]
        elif isinstance(self.hasParameter, list):
            self.hasParameter.append(numerical_variable)
        else:
            self.hasParameter = [self.hasParameter,
                                 numerical_variable]


class Assignment(Thing):
    """not yet implemented"""


class Activity(Thing, abc.ABC):
    """m4i:Activity (not intended to use for modeling)"""


OneOrMultiThings = Union[Thing, HttpUrl, str, List[Union[Thing, HttpUrl, str]]]


@namespaces(m4i="http://w3id.org/nfdi4ing/metadata4ing#",
            schema="https://schema.org/",
            obo="http://purl.obolibrary.org/obo/")
@urirefs(ProcessingStep='m4i:ProcessingStep',
         startTime='schema:startTime',
         endTime='schema:endTime',
         RO_0002224='obo:RO_0002224',
         RO_0002230='obo:RO_0002230',
         hasRuntimeAssignment='m4i:hasRuntimeAssignment',
         investigates='m4i:investigates',
         usageInstruction='m4i:usageInstruction',
         hasEmployedTool='m4i:hasEmployedTool',
         realizesMethod='m4i:realizesMethod',
         hasInput='m4i:hasInput',
         hasOutput='m4i:hasOutput',
         partOf='m4i:partOf',
         precedes='m4i:precedes')
class ProcessingStep(Activity):
    """Pydantic Model for m4i:ProcessingStep

    .. note::

        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    tbd
    """
    startTime: datetime = Field(default=None, alias="start_time")
    endTime: datetime = Field(default=None, alias="end_time")
    RO_0002224: Any = Field(default=None, alias="starts_with")
    RO_0002230: Any = Field(default=None, alias="ends_with")
    hasRuntimeAssignment: Assignment = Field(default=None, alias="runtime_assignment")
    investigates: Thing = None
    usageInstruction: str = Field(default=None, alias="usage_instruction")
    hasEmployedTool: Tool = Field(default=None, alias="has_employed_tool")
    realizesMethod: Union[Method, List[Method]] = Field(default=None, alias="realizes_method")
    hasInput: Thing = Field(default=None, alias="has_input")
    hasOutput: OneOrMultiThings = Field(default=None, alias="has_output")
    partOf: Union[ResearchProject, "ProcessingStep"] = Field(default=None, alias="part_of")
    precedes: "ProcessingStep" = None

    @field_validator('hasOutput', 'hasInput', mode='before')
    @classmethod
    def _one_or_multiple_things(cls, value):
        if isinstance(value, list):
            ret_value = []
            for v in value:
                if isinstance(v, Thing):
                    ret_value.append(v)
                else:
                    if v.startswith("_:"):
                        ret_value.append(v)
                    else:
                        ret_value.append(str(HttpUrl(v)))
            return ret_value
        if isinstance(value, Thing):
            return value
        if str(value).startswith("_:"):
            return value
        return str(HttpUrl(value))

    @field_validator('RO_0002224', mode='before')
    @classmethod
    def _starts_with(cls, starts_with):
        return _validate_processing_step(starts_with)

    @field_validator('RO_0002230', mode='before')
    @classmethod
    def _ends_with(cls, ends_with):
        return _validate_processing_step(ends_with)

    @property
    def starts_with(self):
        return self.RO_0002224

    @starts_with.setter
    def starts_with(self, starts_with):
        self.RO_0002224 = _validate_processing_step(starts_with)

    @property
    def ends_with(self):
        return self.RO_0002230

    @ends_with.setter
    def ends_with(self, ends_with):
        self.RO_0002230 = _validate_processing_step(ends_with)


def _validate_processing_step(ps) -> ProcessingStep:
    if isinstance(ps, ProcessingStep):
        return ps
    if isinstance(ps, dict):
        return ProcessingStep(**ps)
    raise TypeError("starts_with must be of type ProcessingStep or a dictionary")


from ..dcat.resource import Distribution

# add new field to Distribution wasGeneratedBy: ProcessingStep = Field(default=None, alias='was_generated_by'):

Distribution.wasGeneratedBy: ProcessingStep = Field(default=None, alias='was_generated_by')



import pydantic
from pydantic import HttpUrl
from typing import Dict, Type
from . import utils
from .utils import split_URIRef

URIRefManager = utils.UNManager()
NamespaceManager = utils.UNManager()


def _is_http_url(url: str) -> bool:
    """Check if a string is a valid http url.

    Parameters
    ----------
    url : str
        The string to check.

    Returns
    -------
    bool
        True if the string is a valid http url, False otherwise.
    """
    if not str(url).startswith("http"):
        return False
    # now, check for pattern
    try:
        HttpUrl(url)
    except pydantic.ValidationError:
        return False
    return True


def _add_namesapces(cls, namespaces: Dict):
    for k, v in namespaces.items():
        NamespaceManager[cls][k] = str(HttpUrl(v))

def namespaces(**kwargs):
    """Decorator for model classes. It assigns the namespaces used in the uri fields of the class.

    Example:
    --------
    @namespaces(ex="https://example.com/")
    @urirefs(name="ex:name")
    class ExampleModel(ThingModel):
        name: str

    em = ExampleModel(name="test")
    print(em.dump_jsonld())
    # {
    #     "@context": {
    #         "ex": "https://example.com/"
    #     },
    #     "@graph": [
    #         {
    #             "@id": "ex:test",
    #             "ex:name": "test"
    #         }
    #     ]
    # }
    """

    def _decorator(cls):
        """The actual decorator function. It assigns the namespaces to the class."""
        _add_namesapces(cls, kwargs)
        return cls

    return _decorator


_prefix_dict = {
    'http://xmlns.com/foaf/0.1/': 'foaf',
    'https://www.w3.org/ns/prov#': 'prov',
    'https://www.w3.org/1999/02/22-rdf-syntax-ns#': 'rdf',
    'https://schema.org/': 'schema',
    'http://w3id.org/nfdi4ing/metadata4ing#': 'm4i'
}

def _decorate_urirefs(cls, **kwargs):
    fields = list(cls.model_fields.keys())
    fields.append(cls.__name__)

    # add fields to the class
    for k, v in kwargs.items():
        if not isinstance(v, str):
            raise TypeError(f"{v} must be a string, not {type(v)}")
        if _is_http_url(v):
            ns, key = split_URIRef(v)
            prefix = _prefix_dict.get(ns, None)
            if prefix is None:
                URIRefManager[cls][k] = str(v)
            else:
                NamespaceManager[cls][prefix] = str(ns)
                if k not in fields:
                    raise KeyError(f"Field '{k}' not found in {cls.__name__}")
                URIRefManager[cls][k] = f"{prefix}:{key}"
        else:
            if k not in fields:
                raise KeyError(f"Field '{k}' not found in {cls.__name__}")
            URIRefManager[cls][k] = v

def urirefs(**kwargs):
    """decorator for model classes. It assigns the URIRefs to the fields of the class.

    Example:
    --------
    @urirefs(name=URIRef("https://example.com/name"))
    class ExampleModel(ThingModel):
        name: str


    """

    def _decorator(cls):
        """The actual decorator function. It assigns the URIRefs to the fields of the class."""
        _decorate_urirefs(cls, **kwargs)
        return cls

    return _decorator



"""This module contains the query utility functions for the Thing class."""
import json
import logging
import pathlib
from typing import Union, Dict, List, Optional, Type

import rdflib

from .decorator import URIRefManager, NamespaceManager
from .thing import Thing
from .utils import split_URIRef

import re
from collections.abc import Mapping, Iterable

logger = logging.getLogger('ontolutils')


def process_object(
        _id,
        predicate,
        obj: Union[rdflib.URIRef, rdflib.BNode, rdflib.Literal],
        graph,
        add_type):
    """Process the object of a triple."""
    if isinstance(obj, rdflib.Literal):
        logger.debug(f'Object "{obj}" for predicate "{predicate}" is a literal.')
        return str(obj)

    if isinstance(obj, rdflib.BNode):
        logger.debug(f'"{predicate}" has blank node obj! not optimal... difficult to find children...')
        # find children for predicate with blank node obj
        sub_data = {}
        # collection = []
        for (s, p, o) in graph:
            # logger.debug(s, p, o, obj)
            if str(s) == str(obj):
                if isinstance(o, rdflib.Literal):
                    _, key = split_URIRef(p)
                    sub_data[key] = str(o)
                    continue

                if p == rdflib.RDF.first:
                    # first means we have a collection
                    logger.debug(f'Need to find children of first: {o}')
                    # get list:
                    qs = """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT ?item
WHERE {
  ?a rdf:rest*/rdf:first ?item .
}"""
                    list_res = graph.query(qs)

                    _ids = list(set([str(_id[0]) for _id in list_res]))
                    _data = [_query_by_id(graph, _id, add_type) for _id in _ids]
                    return _data

                if p == rdflib.RDF.type and add_type:
                    sub_data["@type"] = str(o)
                else:
                    # may point to another blank node:
                    if isinstance(o, rdflib.BNode):
                        logger.debug(f'"{o}" is a blank node. Need to find children of it...')
                        _, key = split_URIRef(p)
                        if key in sub_data:
                            if isinstance(sub_data[key], list):
                                sub_data[key].append(process_object(_id, p, o, graph, add_type))
                            else:
                                sub_data[key] = [sub_data[key], process_object(_id, p, o, graph, add_type)]
                        else:
                            sub_data[key] = process_object(_id, p, o, graph, add_type)
                    elif str(o).startswith('http'):
                        # it might be a IRI which is defined inside the JSON-LD:
                        _sub_data = process_object(_id, p, o, graph, add_type)
                        if _sub_data:
                            _, key = split_URIRef(p)
                            sub_data[key] = _sub_data
                    else:
                        logger.debug(f'dont know what to do with {p} and {o}')
        if predicate in sub_data:
            return sub_data[predicate]
        return sub_data

    if isinstance(obj, rdflib.URIRef):
        # could be a type definition or a web IRI
        if _is_type_definition(graph, obj):
            logger.debug('points to a type definition inside the data')
            if obj == _id:
                return str(obj)
            return _query_by_id(graph=graph, _id=obj, add_type=True)

    logger.debug(f'"{obj}" for predicate "{predicate}" is a simple data field.')
    return str(obj)


def get_query_string(cls) -> str:
    """Return the query string for the class."""

    def _get_namespace(key):
        ns = URIRefManager[cls].get(key, f'local:{key}')
        if ':' in ns:
            return ns
        return f'{ns}:{key}'

    query_str = f"""
SELECT *
WHERE {{
    ?id a {_get_namespace(cls.__name__)} .
    ?id ?p ?o .
}}"""
    return query_str


def _query_by_id(graph, _id: Union[str, rdflib.URIRef], add_type: bool) -> Dict:
    """Query the graph by the id. Return the data as a dictionary."""
    _sub_query_string = """SELECT DISTINCT ?p ?o WHERE { <%s> ?p ?o }""" % _id
    _sub_res = graph.query(_sub_query_string)
    out = {'@id': str(_id)}
    for binding in _sub_res.bindings:
        predicate = binding['p']
        obj = binding['o']

        if predicate == rdflib.RDF.type:
            if add_type:
                out['@type'] = str(obj)
            continue

        _, key = split_URIRef(predicate)
        if str(_id) == str(obj):
            # would lead to a circular reference. Example for it: "landingPage" and "_id" are the same.
            # in this case, we return the object as a string
            out[key] = str(obj)
        else:
            if key in out:
                if isinstance(out[key], list):
                    out[key].append(process_object(_id, predicate, obj, graph, add_type))
                else:
                    out[key] = [out[key], process_object(_id, predicate, obj, graph, add_type)]
            else:
                out[key] = process_object(_id, predicate, obj, graph, add_type)

    return out


def _is_type_definition(graph, iri: Union[str, rdflib.URIRef]):
    _sub_query_string = """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
SELECT DISTINCT ?o WHERE { <%s> rdf:type ?o }""" % iri
    _sub_res = graph.query(_sub_query_string)
    return len(_sub_res) == 1


def expand_sparql_res(bindings,
                      graph,
                      add_type: bool,
                      add_context: bool) -> Dict:
    """Expand the SPARQL results. Return a dictionary."""
    out = {}
    # n_ = len(bindings)
    for i, binding in enumerate(bindings):
        logger.debug(
            f'Expanding SPARQL results {i + 1}/{len(bindings)}: {binding["?id"]}, {binding["p"]}, {binding["?o"]}.')

        if isinstance(binding['?id'], rdflib.URIRef):
            _id = str(binding['?id'])
        else:
            _id = binding['?id'].n3()
        if _id not in out:
            out[_id] = {}
            if add_context:
                out[_id] = {'@context': {}}
        p = binding['p'].__str__()
        _, predicate = split_URIRef(p)

        if predicate == 'type':
            if add_type:
                out[_id]['@type'] = str(binding['o'])
            continue
        if add_context:
            out[_id]['@context'][predicate] = str(p)

        obj = str(binding['?o'])

        logger.debug(f'Processing object "{obj}" for predicate "{predicate}".')
        data = process_object(_id, predicate, binding['?o'], graph, add_type)

        # well obj is just a data field, add it
        if predicate in out[_id]:
            if isinstance(out[_id][predicate], list):
                out[_id][predicate].append(data)
            else:
                out[_id][predicate] = [out[_id][predicate], data]
        else:
            out[_id][predicate] = data

    return out


def dquery(subject: str,
           source: Optional[Union[str, pathlib.Path]] = None,
           data: Optional[Union[str, Dict]] = None,
           context: Optional[Dict] = None) -> List[Dict]:
    """Return a list of resutls. The entries are dictionaries.

    Example
    -------
    >>> # Query all agents from the source file
    >>> import ontolutils
    >>> ontolutils.dquery(subject='prov:Agent', source='agent1.jsonld')
    """
    g = rdflib.Graph()
    g.parse(source=source,
            data=data,
            format='json-ld',
            context=context)
    if context is None:
        context = {}
    prefixes = "".join([f"PREFIX {k}: <{p}>\n" for k, p in context.items() if not k.startswith('@')])

    assert isinstance(subject, str), f"Subject must be a string, not {type(subject)}"

    query_str = f"""
    SELECT *
    WHERE {{
        ?id a {subject}.
        ?id ?p ?o .
}}"""

    res = g.query(prefixes + query_str)

    if len(res) == 0:
        return []

    logger.debug(f'Querying subject="{subject}" with query: "{prefixes + query_str}" and got {len(res)} results')

    kwargs: Dict = expand_sparql_res(res.bindings, g, True, True)
    for _id in kwargs:
        kwargs[_id]['@id'] = _id
    return [v for v in kwargs.values()]


def query(cls: Type[Thing],
          source: Optional[Union[str, pathlib.Path]] = None,
          data: Optional[Union[str, Dict]] = None,
          context: Optional[Union[Dict, str]] = None,
          limit: Optional[int] = None) -> Union[Thing, List]:
    """Return a generator of results from the query.

    Parameters
    ----------
    cls : Thing
        The class to query
    source: Optional[Union[str, pathlib.Path]]
        The source of the json-ld file. see json.dump() for details
    data : Optional[Union[str, Dict]]
        The data of the json-ld file
    context : Optional[Union[Dict, str]]
        The context of the json-ld file
    limit: Optional[int]
        The limit of the query. Default is None (no limit).
        If limit equals to 1, the result will be a single obj, not a list.
    """
    if cls == Thing:
        query_string = '\nSELECT *\nWHERE {\n    ?id a ?type .\n    ?id ?p ?o .\n}'
    else:
        query_string = get_query_string(
            cls)  # TODO: limit should be passed here, however the sparql query must be written yet
    g = rdflib.Graph()

    ns_keys = [_ns[0] for _ns in g.namespaces()]

    prefixes = "".join([f"PREFIX {k}: <{p}>\n" for k, p in NamespaceManager[cls].items() if not k.startswith('@')])
    for k, p in NamespaceManager[cls].items():
        if k not in ns_keys:
            g.bind(k, p)
            # logger.debug(k)
        g.bind(k, p)

    if isinstance(data, dict):
        data = json.dumps(data)

    _context = cls.get_context()

    if context is None:
        context = {}

    if not isinstance(context, dict):
        raise TypeError(f"Context must be a dict, not {type(context)}")

    _context.update(context)

    g.parse(source=source,
            data=data,
            format='json-ld',
            context=_context)

    gquery = prefixes + query_string

    logger.debug(f'Querying class "{cls.__name__}" with query: {gquery}')
    # logger.debug(prefixes + query_string)
    res = g.query(gquery)

    logger.debug(f'Querying resulted in {len(res)} results')

    if len(res) == 0:
        return []

    logger.debug(f'Expanding SPARQL results...')
    kwargs: Dict = expand_sparql_res(res.bindings, g, False, False)

    # in case that the model field names are different than the IRI names, we need to find the
    # corresponding names. The urirefs translate the class model fields to a <prefix>:<name> format.
    # As we have the latter, the inverse dictionary let's us find the model field names.
    # inverse_urirefs = {_v.split(':', 1)[-1]: _k for _k, _v in get_urirefs(cls).items()}

    kwargs = _at_id_to_id(_expand_compact_iris(kwargs,_context))

    if limit is not None:
        out = []
        for i, (k, v) in enumerate(kwargs.items()):
            model_field_dict = v  # {inverse_urirefs.get(key, key): value for key, value in v.items()}
            if limit == 1:
                return cls.model_validate({'id': k, **model_field_dict})
            out.append(cls.model_validate({'id': k, **model_field_dict}))
            if i == limit - 1:
                return out

    return [cls.model_validate({'id': _id, **params}) for _id, params in kwargs.items()]



_SCHEME_RE = re.compile(r'^[A-Za-z][A-Za-z0-9+.-]*:')

def _expand_compact_iris(obj, context, *, transform_keys=False):
    """
    Recursively expand 'prefix:suffix' strings using a JSON-LD-like context.

    - obj: any nested structure (dict/list/tuple/set/str/…).
    - context: mapping like {'pivmeta': 'https://…#', ...}.
      If a context value is a dict (e.g. {'@id': '…'}), '@id' is used.
    - transform_keys: if True, also expand dictionary KEYS that are 'prefix:suffix'.

    Returns a new structure with expansions applied.
    """
    # Normalize context to {prefix: base_iri_string}
    prefixes = {}
    for k, v in context.items():
        if k.startswith('@'):
            continue
        if isinstance(v, str):
            prefixes[k] = v
        elif isinstance(v, Mapping):
            iri = v.get('@id')
            if isinstance(iri, str):
                prefixes[k] = iri

    def expand_str(s: str) -> str:
        # Only consider exact 'prefix:suffix' forms where prefix is in context.
        if ':' in s and not s.startswith(('http://', 'https://', 'urn:', 'mailto:')):
            pref, rest = s.split(':', 1)
            base = prefixes.get(pref)
            if base:
                # Ensure we don't accidentally drop/add separators incorrectly
                needs_sep = not (base.endswith(('/', '#')) or rest.startswith(('/', '#')))
                sep = '#' if needs_sep else ''
                return f"{base}{sep}{rest}"
        return s

    def expand_any(x):
        if isinstance(x, str):
            return expand_str(x)
        elif isinstance(x, Mapping):
            if transform_keys:
                return {expand_str(k) if isinstance(k, str) else k: expand_any(v)
                        for k, v in x.items()}
            else:
                return {k: expand_any(v) for k, v in x.items()}
        elif isinstance(x, tuple):
            return tuple(expand_any(i) for i in x)
        elif isinstance(x, set):
            return {expand_any(i) for i in x}
        elif isinstance(x, list):
            return [expand_any(i) for i in x]
        else:
            return x

    return expand_any(obj)


def _at_id_to_id(data: Dict) -> Dict:
    """Convert recursively '@id' keys to 'id' keys in a dictionary."""
    """Convert '@id' keys to 'id' keys in a dictionary."""
    if isinstance(data, dict):
        new_data = {}
        for k, v in data.items():
            if k == '@id':
                new_data['id'] = v
            else:
                new_data[k] = _at_id_to_id(v)
        return new_data
    elif isinstance(data, list):
        return [_at_id_to_id(item) for item in data]
    else:
        return data